//! Public or symmetric key wrappers, nonces and all that magic stuff

pub mod asym;
mod errors;
pub mod hkdf;
pub mod sym;

pub use errors::Error;

use ::ring::rand::SecureRandom;

/// wrapper where we implement whatever random traint stuff each library needs
pub struct Random {
    /// actual source of randomness
    rnd: ::ring::rand::SystemRandom,
}

impl Random {
    /// Build a nre Random source
    pub fn new() -> Self {
        Self {
            rnd: ::ring::rand::SystemRandom::new(),
        }
    }
    /// Fill a buffer with randomness
    pub fn fill(&self, out: &mut [u8]) {
        self.rnd.fill(out);
    }
}

// Fake debug implementation to avoid leaking secrets
impl ::core::fmt::Debug for Random {
    fn fmt(
        &self,
        f: &mut core::fmt::Formatter<'_>,
    ) -> Result<(), ::std::fmt::Error> {
        ::core::fmt::Debug::fmt("[hidden randomness]", f)
    }
}

// ::rand_core::{RngCore, CryptoRng} needed for ::x25519::dalek
impl ::rand_core::RngCore for &Random {
    fn next_u32(&mut self) -> u32 {
        use ::core::mem::MaybeUninit;
        let mut out: MaybeUninit<[u8; 4]> = MaybeUninit::uninit();
        #[allow(unsafe_code)]
        unsafe {
            let _ = self.rnd.fill(out.assume_init_mut());
            u32::from_le_bytes(out.assume_init())
        }
    }
    fn next_u64(&mut self) -> u64 {
        use ::core::mem::MaybeUninit;
        let mut out: MaybeUninit<[u8; 8]> = MaybeUninit::uninit();
        #[allow(unsafe_code)]
        unsafe {
            let _ = self.rnd.fill(out.assume_init_mut());
            u64::from_le_bytes(out.assume_init())
        }
    }
    fn fill_bytes(&mut self, dest: &mut [u8]) {
        let _ = self.rnd.fill(dest);
    }
    fn try_fill_bytes(
        &mut self,
        dest: &mut [u8],
    ) -> Result<(), ::rand_core::Error> {
        match self.rnd.fill(dest) {
            Ok(()) => Ok(()),
            Err(e) => Err(::rand_core::Error::new(e)),
        }
    }
}
impl ::rand_core::CryptoRng for &Random {}
